// #include <cassert>
#include <chrono>
#include <cublas_v2.h>
#include <cudnn.h>
#include <iostream>
#include <sstream>
#include <string.h>
#include <time.h>
#include <unordered_map>
#include <vector>
#include "BatchStreamPPM.h"
#include "NvUffParser.h"
#include "/home/ets-zlc/TensorRT-5.1.5.0.Ubuntu-18.04.2.x86_64-gnu.cuda-10.0.cudnn7.5/TensorRT-5.1.5.0/samples/common/common.h"
#include "/home/ets-zlc/TensorRT-5.1.5.0.Ubuntu-18.04.2.x86_64-gnu.cuda-10.0.cudnn7.5/TensorRT-5.1.5.0/samples/common/argsParser.h"
#include "NvInferPlugin.h"

#define ENABLE_LOGS 1
#define OPTION_1 0
using namespace nvinfer1;
using namespace nvuffparser;

Logger gLogger;
samplesCommon::Args args;

#define RETURN_AND_LOG(ret, severity, message)                                 \
    do                                                                         \
    {                                                                          \
        std::string error_message = "MYEXP: " + std::string(message); \
        gLogger.log(ILogger::Severity::k##severity, error_message.c_str());    \
        return (ret);                                                          \
    } while (0)


const char* OUTPUT_BLOB_NAME0 = "Conv2D";

int outH = 181;
int outW = 321;
int outC = 32;


//INT8 Calibration, currently set to calibrate over 500 images
static constexpr int CAL_BATCH_SIZE = 50;
static constexpr int FIRST_CAL_BATCH = 0, NB_CAL_BATCHES = 10;

// Visualization
const float visualizeThreshold = 0.5;

std::string locateFile(const std::string& input)
{
	std::vector<std::string> dirs{ "../../tensorflowcode/pre-trained" };
	return locateFile(input, dirs);
}


std::vector<std::pair<int64_t, DataType>>
calculateBindingBufferSizes(const ICudaEngine& engine, int nbBindings, int batchSize)
{
	std::vector<std::pair<int64_t, DataType>> sizes;
	for (int i = 0; i < nbBindings; ++i)
	{
		Dims dims = engine.getBindingDimensions(i);
		DataType dtype = engine.getBindingDataType(i);
#if ENABLE_LOGS
		const char  *name = engine.getBindingName(i);
		std::cout << "for binding" << i << "\n";
		std::cout << name << "   ";
		for (int j = 0;j < dims.nbDims;j++)
		{
			std::cout << "x" << dims.d[j];
		}
#endif
		int64_t eltCount = samplesCommon::volume(dims) * batchSize;
		sizes.push_back(std::make_pair(eltCount, dtype));
	}

	return sizes;
}

ICudaEngine* loadModelAndCreateEngine(const char* uffFile, int maxBatchSize,
	IUffParser* parser, nvuffparser::IPluginFactory* pluginFactory, IInt8Calibrator* calibrator, IHostMemory*& trtModelStream)
{
	// Create the builder
	IBuilder* builder = createInferBuilder(gLogger);

	// Parse the UFF model to populate the network, then set the outputs.
	INetworkDefinition* network = builder->createNetwork();
	//set the plugin factory
	//parser->setPluginFactory(pluginFactory);
#if ENABLE_LOGS
	std::cout << "Begin parsing model..." << std::endl;
#endif
	if (!parser->parse(uffFile, *network, nvinfer1::DataType::kFLOAT))
		RETURN_AND_LOG(nullptr, ERROR, "Fail to parse");
#if ENABLE_LOGS
	std::cout << "End parsing model..." << std::endl;
#endif
	// Build the engine.
	builder->setMaxBatchSize(maxBatchSize);
	// The _GB literal operator is defined in common/common.h
	builder->setMaxWorkspaceSize(1_GB); // We need about 1GB of scratch space for the plugin layer for batch size 5.
	builder->setHalf2Mode(false);
	if (args.runInInt8)
	{
		builder->setInt8Mode(true);
		builder->setInt8Calibrator(calibrator);
	}
#if ENABLE_LOGS
	std::cout << "Begin building engine..." << std::endl;
#endif
	ICudaEngine* engine = builder->buildCudaEngine(*network);
	if (!engine)
		RETURN_AND_LOG(nullptr, ERROR, "Unable to create engine");
#if ENABLE_LOGS
	std::cout << "End building engine..." << std::endl;
#endif
	// We don't need the network any more, and we can destroy the parser.
	network->destroy();
	parser->destroy();

	// Serialize the engine, then close everything down.
	trtModelStream = engine->serialize();

	builder->destroy();
	shutdownProtobufLibrary();
	return engine;
}

void doInference(IExecutionContext& context, float* inputData, float* detectionOut, int* keepCount, int batchSize)
{
	const ICudaEngine& engine = context.getEngine();
	int nbBindings = engine.getNbBindings();

	std::vector<void*> buffers(nbBindings);
	std::vector<std::pair<int64_t, DataType>> buffersSizes = calculateBindingBufferSizes(engine, nbBindings, batchSize);

	for (int i = 0; i < nbBindings; ++i)
	{
		auto bufferSizesOutput = buffersSizes[i];
		buffers[i] = samplesCommon::safeCudaMalloc(bufferSizesOutput.first * samplesCommon::getElementSize(bufferSizesOutput.second));
	}

	// In order to bind the buffers, we need to know the names of the input and output tensors.
	// Note that indices are guaranteed to be less than IEngine::getNbBindings().
	int inputIndex = engine.getBindingIndex(INPUT_BLOB_NAME),
		outputIndex0 = engine.getBindingIndex(OUTPUT_BLOB_NAME0);

	cudaStream_t stream;
	CHECK(cudaStreamCreate(&stream));

	// DMA the input to the GPU,  execute the batch asynchronously, and DMA it back:
	CHECK(cudaMemcpyAsync(buffers[inputIndex], inputData, batchSize * INPUT_C * INPUT_H * INPUT_W * sizeof(float), cudaMemcpyHostToDevice, stream));

	auto t_start = std::chrono::high_resolution_clock::now();
	context.execute(batchSize, &buffers[0]);
	auto t_end = std::chrono::high_resolution_clock::now();
	float total = std::chrono::duration<float, std::milli>(t_end - t_start).count();

	std::cout << "Time taken for inference is " << total << " ms." << std::endl;

	for (int bindingIdx = 0; bindingIdx < nbBindings; ++bindingIdx)
	{
		if (engine.bindingIsInput(bindingIdx))
			continue;
	}

	CHECK(cudaMemcpyAsync(detectionOut, buffers[outputIndex0], batchSize * outH*outW*outC * sizeof(float), cudaMemcpyDeviceToHost, stream));
	// FORGET COPYING OUTPUT TILL NO ISSUE IN EXEC
	// CHECK(cudaMemcpyAsync(keepCount, buffers[outputIndex1], batchSize * sizeof(int), cudaMemcpyDeviceToHost, stream));
	cudaStreamSynchronize(stream);
	// Release the stream and the buffers
	cudaStreamDestroy(stream);
	CHECK(cudaFree(buffers[inputIndex]));
	CHECK(cudaFree(buffers[outputIndex0]));
}

class pad : public IPluginV2
{
public:
	pad(int in_wPad, int  in_hPad)
		:wPad(in_wPad), hPad(in_hPad)
	{
	}
	pad(const void* data, size_t length)
	{
		const char *d = reinterpret_cast<const char *>(data), *a = d;
		for (int i = 0; i < 6; ++i)
		{
			mCopySize[i] = read<int>(d);
		}
		wPad = read<int>(d);
		hPad = read<int>(d);
	}
	~pad()
	{

	}
	int getNbOutputs() const override {
		return 1;
	}
	Dims getOutputDimensions(int index, const Dims* inputs, int nbInputDims) override
	{
#if OPTION_1
		std::cout << " Inside get getOutputDimensions\n" << inputs[0].d[0] << " " << inputs[0].d[1] << " " << inputs[0].d[2]<<std::endl;
		// PAD BY 6
		return DimsCHW(inputs[0].d[0], inputs[0].d[1] + 6, inputs[0].d[2] + 6);
#else
		std::cout << " Inside get getOutputDimensions\n" << inputs[0].d[0] << " " << inputs[0].d[1] << " " << inputs[0].d[2] << std::endl;
		// PAD BY 6
		return DimsNCHW(1,inputs[0].d[0], inputs[0].d[1] + 6, inputs[0].d[2] + 6);

#endif
	}
	int initialize() override
	{
		return 0;
	}
	void terminate() override
	{
	}
	size_t getWorkspaceSize(int) const override {
		return 0;
	}

	int enqueue(int batchSize, const void* const* inputs, void** outputs, void*, cudaStream_t stream) override
	{
		std::cout << "inside enqueue\n";
		int in_channels = mCopySize[0];
		int in_height = mCopySize[1];
		int in_width = mCopySize[2];
		int out_channels = mCopySize[3];
		int out_height = mCopySize[4];
		int out_width = mCopySize[5];

		const float* input = reinterpret_cast<const float*>(inputs[0]);
		float* output = reinterpret_cast<float*>(outputs[0]);
		int input_size = in_channels * in_height * in_width * sizeof(float);
		std::cout << "at line 419---\n" << input_size << " " << in_channels << " " << in_height << " " << in_width << " ";

		// allocate the memory
		float* inputTemp;
		CHECK(cudaMalloc(&inputTemp, input_size));
		//memcpy data
		CHECK(cudaMemcpy(inputTemp, input, input_size, cudaMemcpyDeviceToDevice));
		// CODE : TODO
		//Pad_zeros(inputTemp, output, mCopySize, wPad, hPad);
		float* output_c = (float*)malloc(out_channels* out_height * out_width * sizeof(float));
		CHECK(cudaMemcpy(output_c, output, out_channels* out_height * out_width * sizeof(float), cudaMemcpyDeviceToHost));
	}
	size_t getSerializationSize() const override
	{
		return 8 * sizeof(int);
	}
	void serialize(void* buffer) const override
	{
		char *d = reinterpret_cast<char *>(buffer), *a = d;
		for (int i = 0; i < 6; ++i)
		{
			write(d, mCopySize[i]);
		}
		write(d, wPad);
		write(d, hPad);
		assert(d == a + getSerializationSize());
	}
	void configureWithFormat(const Dims* inputs, int nbInputs, const Dims* outputDims, int nbOutputs, nvinfer1::DataType type, nvinfer1::PluginFormat format, int maxBatchSize) override
	{

		// save  the input and ouptut dimensions
		mCopySize[0] = inputs[0].d[0];
		mCopySize[1] = inputs[0].d[1];
		mCopySize[2] = inputs[0].d[2];
#if OPTION_1
		mCopySize[3] = outputDims[0].d[0];
		mCopySize[4] = outputDims[0].d[1];
		mCopySize[5] = outputDims[0].d[2];
#else
		mCopySize[3] = outputDims[0].d[1];
		mCopySize[4] = outputDims[0].d[2];
		mCopySize[5] = outputDims[0].d[3];

#endif

	}
	bool supportsFormat(DataType type, PluginFormat format) const override
	{
		return (type == DataType::kFLOAT && format == PluginFormat::kNCHW);
	}
	const char* getPluginType() const override { return "Pad_Custom"; }
	const char* getPluginVersion() const override { return "1"; }
	void destroy() override { delete this; }
	IPluginV2* clone() const override
	{
		return new pad(3, 3);
	}
	void setPluginNamespace(const char* libNamespace) override { mNamespace = libNamespace; }
	const char* getPluginNamespace() const override { return mNamespace.c_str(); }
private:
	int mCopySize[6];
	template <typename T>
	void write(char*& buffer, const T& val) const
	{
		*reinterpret_cast<T*>(buffer) = val;
		buffer += sizeof(T);
	}
	template <typename T>
	T read(const char*& buffer)
	{
		T val = *reinterpret_cast<const T*>(buffer);
		buffer += sizeof(T);
		return val;
	}

	bool mIgnoreBatch{ false };
	int mConcatAxisID{ 0 }, mOutputConcatAxis{ 0 }, mNumInputs{ 0 };
	int* mInputConcatAxis = nullptr;
	nvinfer1::Dims mCHW;
	cublasHandle_t mCublas;
	std::string mNamespace;
	int wPad{ 0 }, hPad{ 0 };
};
namespace
{
	const char* pad_PLUGIN_VERSION{ "1" };
	const char* pad_PLUGIN_NAME{ "Pad_Custom" };
} // namespace
class padPluginCreator : public IPluginCreator
{
public:
	padPluginCreator()
	{
		mFC.nbFields = mPluginAttributes.size();
		mFC.fields = mPluginAttributes.data();
	}
	~padPluginCreator() {}
	const char* getPluginName() const override { return pad_PLUGIN_NAME; }
	const char* getPluginVersion() const override { return pad_PLUGIN_VERSION; }
	const PluginFieldCollection* getFieldNames() override { return &mFC; }
	IPluginV2* createPlugin(const char* name, const PluginFieldCollection* fc) override
	{
		const PluginField* fields = fc->fields;
		return new pad(3, 3);
	}
	IPluginV2* deserializePlugin(const char* name, const void* serialData, size_t serialLength) override
	{
		return new pad(serialData, serialLength);
	}
	void setPluginNamespace(const char* libNamespace) override { mNamespace = libNamespace; }
	const char* getPluginNamespace() const override { return mNamespace.c_str(); }
private:
	static PluginFieldCollection mFC;
	static std::vector<PluginField> mPluginAttributes;
	std::string mNamespace = "";
};
PluginFieldCollection padPluginCreator::mFC{};
std::vector<PluginField> padPluginCreator::mPluginAttributes;
REGISTER_TENSORRT_PLUGIN(padPluginCreator);

int main(int argc, char* argv[])
{
	// Parse command-line arguments.
	samplesCommon::parseArgs(args, argc, argv);

	initLibNvInferPlugins(&gLogger, "");

	auto fileName = locateFile("frozen_model.uff");

	std::cout << fileName << std::endl;

	const int N = 1;
	auto parser = createUffParser();

	gLogger.setReportableSeverity(Severity::kINFO);
	//BatchStream calibrationStream(CAL_BATCH_SIZE, NB_CAL_BATCHES);

	parser->registerInput("Input", DimsNCHW(8, 3, 361, 641), UffInputOrder::kNCHW);
	parser->registerOutput("transpose");

	IHostMemory* trtModelStream{ nullptr };
	ICudaEngine* tmpEngine = loadModelAndCreateEngine(fileName.c_str(), N, parser, NULL, NULL, trtModelStream);
	assert(tmpEngine != nullptr);
	assert(trtModelStream != nullptr);
	tmpEngine->destroy();

	// Read a random sample image.
	srand(unsigned(time(nullptr)));

	// Available images.
	std::vector<std::string> imageList = { "input.ppm" };
	std::vector<samplesCommon::PPM<INPUT_C, INPUT_H, INPUT_W>> ppms(N);

	assert(ppms.size() <= imageList.size());
	for (int i = 0; i < N; ++i)
	{
		//readPPMFile(locateFile(imageList[i]), ppms[i]);
	}

	vector<float> data(N * INPUT_C * INPUT_H * INPUT_W);
	//TODO: DONT LOAD DATA NOW

	//PluginFactory pluginFactory;
	// Deserialize the engine.
	IRuntime* runtime = createInferRuntime(gLogger);
	assert(runtime != nullptr);
	ICudaEngine* engine = runtime->deserializeCudaEngine(trtModelStream->data(), trtModelStream->size(), NULL);
	assert(engine != nullptr);
	trtModelStream->destroy();
	IExecutionContext* context = engine->createExecutionContext();
	assert(context != nullptr);

	// Host memory for outputs.
	vector<float> detectionOut(N * outC*outW*outH);
	vector<int> keepCount(N);
	// Run inference.
	doInference(*context, &data[0], &detectionOut[0], &keepCount[0], N);

	context->destroy();
	engine->destroy();
	runtime->destroy();

	return EXIT_SUCCESS;
}

